import os
import pathlib
import time
from datetime import datetime
from unittest import mock

import pandas as pd
import pytest
from pyspark.sql import SparkSession

from mlflow.exceptions import MlflowException
from mlflow.pipelines.steps.ingest import IngestStep
from mlflow.store.artifact.s3_artifact_repo import S3ArtifactRepository

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    enter_pipeline_example_directory,
    enter_test_pipeline_directory,
)


@pytest.fixture
def pandas_df():
    df = pd.DataFrame.from_dict(
        {
            "A": ["x", "y", "z"],
            "B": [1, 2, 3],
            "C": [-9.2, 82.5, 3.40],
        }
    )
    df.index.rename("index", inplace=True)
    return df


@pytest.fixture(scope="module", autouse=True)
def spark_session():
    session = (
        SparkSession.builder.master("local[*]")
        .config("spark.jars.packages", "io.delta:delta-core_2.12:1.2.1")
        .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
        .config(
            "spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog"
        )
        .getOrCreate()
    )
    yield session
    session.stop()


@pytest.fixture()
def spark_df(spark_session):
    return spark_session.createDataFrame(
        [
            (0, "a b c d e spark", 1.0),
            (1, "b d", 0.0),
            (2, "spark f g h", 1.0),
            (3, "hadoop mapreduce", 0.0),
        ],
        ["id", "text", "label"],
    ).cache()


@pytest.mark.parametrize("use_relative_path", [False, True])
@pytest.mark.parametrize("multiple_files", [False, True])
@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_parquet_successfully(use_relative_path, multiple_files, pandas_df, tmp_path):
    if multiple_files:
        dataset_path = tmp_path / "dataset"
        dataset_path.mkdir()
        pandas_df_part1 = pandas_df[:1]
        pandas_df_part2 = pandas_df[1:]
        pandas_df_part1.to_parquet(dataset_path / "df1.parquet")
        pandas_df_part2.to_parquet(dataset_path / "df2.parquet")
    else:
        dataset_path = tmp_path / "df.parquet"
        pandas_df.to_parquet(dataset_path)

    if use_relative_path:
        dataset_path = os.path.relpath(dataset_path)

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": str(dataset_path),
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    reloaded_df = pd.read_parquet(str(tmp_path / "dataset.parquet"))
    pd.testing.assert_frame_equal(reloaded_df, pandas_df)


@pytest.mark.parametrize("use_relative_path", [False, True])
@pytest.mark.parametrize("multiple_files", [False, True])
@pytest.mark.parametrize("explicit_file_list", [False, True])
@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_csv_successfully(
    use_relative_path,
    multiple_files,
    explicit_file_list,
    pandas_df,
    tmp_path,
):
    if multiple_files:
        dataset_path = tmp_path / "dataset"
        dataset_path.mkdir()
        pandas_df_part1 = pandas_df[:1]
        pandas_df_part2 = pandas_df[1:]
        pandas_df_part1.to_csv(dataset_path / "df1.csv")
        pandas_df_part2.to_csv(dataset_path / "df2.csv")
    else:
        dataset_path = tmp_path / "df.csv"
        pandas_df.to_csv(dataset_path)

    if use_relative_path:
        dataset_path = pathlib.Path(os.path.relpath(dataset_path))

    if explicit_file_list and multiple_files:
        dataset_path = [f'{dataset_path / "df1.csv"}', f'{dataset_path / "df2.csv"}']
    else:
        dataset_path = str(dataset_path)

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "csv",
                "location": dataset_path,
                "custom_loader_method": "steps.ingest.load_file_as_dataframe",
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    reloaded_df = pd.read_parquet(str(tmp_path / "dataset.parquet"))
    pd.testing.assert_frame_equal(reloaded_df, pandas_df)


def custom_load_wine_csv(file_path, file_format):  # pylint: disable=unused-argument
    return pd.read_csv(file_path, sep=";")


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_remote_http_datasets_with_multiple_files_successfully(tmp_path):
    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "csv",
                "location": [
                    "https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-red.csv",
                    "https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-white.csv",
                ],
                "custom_loader_method": "tests.pipelines.test_ingest_step.custom_load_wine_csv",
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)
    reloaded_df = pd.read_parquet(str(tmp_path / "dataset.parquet"))
    assert reloaded_df.count()[0] == 6497


def custom_load_file_as_dataframe(file_path, file_format):  # pylint: disable=unused-argument
    return pd.read_csv(file_path, sep="#", index_col=0)


@pytest.mark.parametrize("use_relative_path", [False, True])
@pytest.mark.parametrize("multiple_files", [False, True])
@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_custom_format_successfully(use_relative_path, multiple_files, pandas_df, tmp_path):
    if multiple_files:
        dataset_path = tmp_path / "dataset"
        dataset_path.mkdir()
        pandas_df_part1 = pandas_df[:1]
        pandas_df_part2 = pandas_df[1:]
        pandas_df_part1.to_csv(dataset_path / "df1.fooformat", sep="#")
        pandas_df_part2.to_csv(dataset_path / "df2.fooformat", sep="#")
    else:
        dataset_path = tmp_path / "df.fooformat"
        pandas_df.to_csv(dataset_path, sep="#")

    if use_relative_path:
        dataset_path = os.path.relpath(dataset_path)

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "fooformat",
                "location": str(dataset_path),
                "custom_loader_method": (
                    "tests.pipelines.test_ingest_step.custom_load_file_as_dataframe"
                ),
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    reloaded_df = pd.read_parquet(str(tmp_path / "dataset.parquet"))
    pd.testing.assert_frame_equal(reloaded_df, pandas_df)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_for_custom_dataset_when_custom_loader_function_cannot_be_imported(
    pandas_df, tmp_path
):
    dataset_path = tmp_path / "df.fooformat"
    pandas_df.to_csv(dataset_path, sep="#")

    with pytest.raises(MlflowException, match="Failed to import custom dataset loader function"):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "format": "fooformat",
                    "location": str(dataset_path),
                    "custom_loader_method": "non.existent.module.non.existent.method",
                }
            },
            pipeline_root=os.getcwd(),
        ).run(output_directory=tmp_path)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_for_custom_dataset_when_custom_loader_function_not_implemented_for_format(
    pandas_df, tmp_path
):
    dataset_path = tmp_path / "df.fooformat"
    pandas_df.to_csv(dataset_path, sep="#")

    with pytest.raises(
        MlflowException, match="Please update the custom loader method to support this format"
    ):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "format": "fooformat",
                    "location": str(dataset_path),
                    "custom_loader_method": "steps.ingest.load_file_as_dataframe",
                }
            },
            pipeline_root=os.getcwd(),
        ).run(output_directory=tmp_path)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_for_custom_dataset_when_custom_loader_function_throws_unexpectedly(
    pandas_df, tmp_path
):
    dataset_path = tmp_path / "df.fooformat"
    pandas_df.to_csv(dataset_path, sep="#")

    with mock.patch(
        "tests.pipelines.test_ingest_step.custom_load_file_as_dataframe",
        side_effect=Exception("Failed to load!"),
    ) as mock_custom_loader:
        setattr(mock_custom_loader, "__name__", "custom_load_file_as_dataframe")
        with pytest.raises(
            MlflowException, match="Unable to load data file at path.*using custom loader method"
        ):
            IngestStep.from_pipeline_config(
                pipeline_config={
                    "data": {
                        "format": "fooformat",
                        "location": str(dataset_path),
                        "custom_loader_method": (
                            "tests.pipelines.test_ingest_step.custom_load_file_as_dataframe"
                        ),
                    }
                },
                pipeline_root=os.getcwd(),
            ).run(output_directory=tmp_path)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_remote_s3_datasets_successfully(mock_s3_bucket, pandas_df, tmp_path):
    dataset_path = tmp_path / "df.parquet"
    pandas_df.to_parquet(dataset_path)
    S3ArtifactRepository(f"s3://{mock_s3_bucket}").log_artifact(str(dataset_path))

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": f"s3://{mock_s3_bucket}/df.parquet",
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    reloaded_df = pd.read_parquet(str(tmp_path / "dataset.parquet"))
    pd.testing.assert_frame_equal(reloaded_df, pandas_df)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_remote_http_datasets_successfully(tmp_path):
    dataset_url = "https://raw.githubusercontent.com/mlflow/mlflow/594a08f2a49c5754bb65d76cd719c15c5b8266e9/examples/sklearn_elasticnet_wine/wine-quality.csv"
    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "csv",
                "location": dataset_url,
                "custom_loader_method": "steps.ingest.load_file_as_dataframe",
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    reloaded_df = pd.read_parquet(str(tmp_path / "dataset.parquet"))
    pd.testing.assert_frame_equal(reloaded_df, pd.read_csv(dataset_url, index_col=0))


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_spark_sql_successfully(spark_df, tmp_path):
    spark_df.write.mode("overwrite").saveAsTable("test_table")

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "spark_sql",
                "sql": "SELECT * FROM test_table ORDER BY id",
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    # Spark DataFrames are not ingested with a consistent row order, as doing so would incur a
    # substantial performance cost. Accordingly, we sort the ingested DataFrame and the original
    # DataFrame on the `id` column and reset the DataFrame index to achieve a consistent ordering
    # before testing their equivalence
    reloaded_df = (
        pd.read_parquet(str(tmp_path / "dataset.parquet"))
        .sort_values(by="id")
        .reset_index(drop=True)
    )
    spark_to_pandas_df = spark_df.toPandas().sort_values(by="id").reset_index(drop=True)
    pd.testing.assert_frame_equal(reloaded_df, spark_to_pandas_df)


@pytest.mark.parametrize("use_relative_path", [False, True])
@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingests_delta_successfully(use_relative_path, spark_df, tmp_path):
    dataset_path = tmp_path / "test.delta"
    spark_df.write.format("delta").save(str(dataset_path))
    if use_relative_path:
        dataset_path = os.path.relpath(dataset_path)

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "delta",
                "location": str(dataset_path),
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    # Spark DataFrames are not ingested with a consistent row order, as doing so would incur a
    # substantial performance cost. Accordingly, we sort the ingested DataFrame and the original
    # DataFrame on the `id` column and reset the DataFrame index to achieve a consistent ordering
    # before testing their equivalence
    reloaded_df = (
        pd.read_parquet(str(tmp_path / "dataset.parquet"))
        .sort_values(by="id")
        .reset_index(drop=True)
    )
    spark_to_pandas_df = spark_df.toPandas().sort_values(by="id").reset_index(drop=True)
    pd.testing.assert_frame_equal(reloaded_df, spark_to_pandas_df)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
@pytest.mark.parametrize("version", [0, 1])
def test_ingests_delta_with_table_version_successfully(spark_session, spark_df, tmp_path, version):
    dataset_path = tmp_path / "test.delta"
    v0_df = spark_df
    v1_df = spark_session.createDataFrame(
        [
            (0, "new df row 0", 10.0),
        ],
        ["id", "text", "label"],
    )
    v0_df.write.format("delta").save(str(dataset_path))
    v1_df.write.format("delta").mode("overwrite").save(str(dataset_path))

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "delta",
                "location": str(dataset_path),
                "version": version,
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    # Spark DataFrames are not ingested with a consistent row order, as doing so would incur a
    # substantial performance cost. Accordingly, we sort the ingested DataFrame and the original
    # DataFrame on the `id` column and reset the DataFrame index to achieve a consistent ordering
    # before testing their equivalence
    reloaded_df = (
        pd.read_parquet(str(tmp_path / "dataset.parquet"))
        .sort_values(by="id")
        .reset_index(drop=True)
    )
    expected_spark_df = v0_df if version == 0 else v1_df
    spark_to_pandas_df = expected_spark_df.toPandas().sort_values(by="id").reset_index(drop=True)
    pd.testing.assert_frame_equal(reloaded_df, spark_to_pandas_df)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
@pytest.mark.parametrize("timestamp_idx", [0, 1])
def test_ingests_delta_with_timestamp_successfully(
    spark_session, spark_df, tmp_path, timestamp_idx
):
    dataset_path = tmp_path / "test.delta"
    v0_df = spark_df
    v1_df = spark_session.createDataFrame(
        [
            (0, "new df row 0", 10.0),
        ],
        ["id", "text", "label"],
    )
    v0_df.write.format("delta").save(str(dataset_path))
    timestamps = [datetime.now().isoformat()]
    time.sleep(2)
    v1_df.write.format("delta").mode("overwrite").save(str(dataset_path))
    timestamps.append(datetime.now().isoformat())
    # Wait a couple of seconds and perform extra write so that all computed timestamps are
    # guaranteed to lie within the range of delta table write times (otherwise, Delta throws a
    # "timestamp out of range" error)
    time.sleep(2)
    spark_session.createDataFrame(
        [
            (0, "row for dummy write", 10.0),
        ],
        ["id", "text", "label"],
    ).write.format("delta").mode("overwrite").save(str(dataset_path))

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "delta",
                "location": str(dataset_path),
                "timestamp": timestamps[timestamp_idx],
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    # Spark DataFrames are not ingested with a consistent row order, as doing so would incur a
    # substantial performance cost. Accordingly, we sort the ingested DataFrame and the original
    # DataFrame on the `id` column and reset the DataFrame index to achieve a consistent ordering
    # before testing their equivalence
    reloaded_df = (
        pd.read_parquet(str(tmp_path / "dataset.parquet"))
        .sort_values(by="id")
        .reset_index(drop=True)
    )
    expected_spark_df = v0_df if timestamp_idx == 0 else v1_df
    spark_to_pandas_df = expected_spark_df.toPandas().sort_values(by="id").reset_index(drop=True)
    pd.testing.assert_frame_equal(reloaded_df, spark_to_pandas_df)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_directory_ignores_files_that_do_not_match_dataset_format(pandas_df, tmp_path):
    dataset_path = tmp_path / "dataset"
    dataset_path.mkdir()
    pandas_df_part1 = pandas_df[:1]
    pandas_df_part2 = pandas_df[1:]
    pandas_df_part1.to_parquet(dataset_path / "df1.parquet")
    pandas_df_part2.to_parquet(dataset_path / "df2.parquet")
    # Ingest should ignore these files
    pandas_df_part1.to_csv(dataset_path / "df1.csv")
    with open(dataset_path / "README", "w") as f:
        f.write("Interesting README content")

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": str(dataset_path),
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    reloaded_df = pd.read_parquet(str(tmp_path / "dataset.parquet"))
    pd.testing.assert_frame_equal(reloaded_df, pandas_df)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_produces_expected_step_card(pandas_df, tmp_path):
    dataset_path = tmp_path / "df.parquet"
    pandas_df.to_parquet(dataset_path)

    IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": str(dataset_path),
            }
        },
        pipeline_root=os.getcwd(),
    ).run(output_directory=tmp_path)

    expected_step_card_path = os.path.join(tmp_path, "card.html")
    assert os.path.exists(expected_step_card_path)
    with open(expected_step_card_path, "r") as f:
        step_card_html_content = f.read()

    assert "Dataset source location" in step_card_html_content
    assert "Number of rows ingested" in step_card_html_content
    assert "Profile of Ingested Dataset" in step_card_html_content


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_when_spark_unavailable_for_spark_based_dataset(spark_df, tmp_path):
    dataset_path = tmp_path / "test.delta"
    spark_df.write.format("delta").save(str(dataset_path))

    with mock.patch(
        "mlflow.pipelines.steps.ingest.datasets._get_active_spark_session",
        side_effect=Exception("Spark unavailable"),
    ), pytest.raises(
        MlflowException, match="Encountered an error while searching for an active Spark session"
    ):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "format": "delta",
                    "location": str(dataset_path),
                }
            },
            pipeline_root=os.getcwd(),
        ).run(output_directory=tmp_path)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_when_dataset_format_unspecified():
    with pytest.raises(MlflowException, match="Dataset format must be specified"):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "location": "my_location",
                }
            },
            pipeline_root=os.getcwd(),
        )


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_when_data_section_unspecified():
    with pytest.raises(MlflowException, match="The `data` section.*must be specified"):
        IngestStep.from_pipeline_config(
            pipeline_config={},
            pipeline_root=os.getcwd(),
        )


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_when_required_dataset_config_keys_are_missing():
    with pytest.raises(MlflowException, match="The `location` configuration key must be specified"):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "format": "parquet",
                    # Missing location
                }
            },
            pipeline_root=os.getcwd(),
        )

    with pytest.raises(MlflowException, match="The `sql` configuration key must be specified"):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "format": "spark_sql",
                    # Missing sql
                }
            },
            pipeline_root=os.getcwd(),
        )

    with pytest.raises(
        MlflowException, match="The `custom_loader_method` configuration key must be specified"
    ):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "format": "csv",
                    "location": "my/dataset.csv",
                    # Missing custom_loader_method
                }
            },
            pipeline_root=os.getcwd(),
        )


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_throws_when_dataset_files_have_wrong_format(pandas_df, tmp_path):
    dataset_path = tmp_path / "df.csv"
    pandas_df.to_csv(dataset_path)

    with pytest.raises(
        MlflowException, match="Resolved data file.*does not have the expected format"
    ):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    # Intentionally use an incorrect format that doesn't match the dataset
                    "format": "parquet",
                    "location": str(dataset_path),
                }
            },
            pipeline_root=os.getcwd(),
        ).run(output_directory=tmp_path)

    dataset_path = tmp_path / "dataset"
    dataset_path.mkdir()
    pandas_df_part1 = pandas_df[:1]
    pandas_df_part2 = pandas_df[1:]
    pandas_df_part1.to_csv(dataset_path / "df1.csv")
    pandas_df_part2.to_csv(dataset_path / "df2.csv")

    with pytest.raises(
        MlflowException, match="Did not find any data files with the specified format"
    ):
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    # Intentionally use an incorrect format that doesn't match the dataset
                    "format": "parquet",
                    "location": str(dataset_path),
                }
            },
            pipeline_root=os.getcwd(),
        ).run(output_directory=tmp_path)


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_ingest_skips_profiling_when_specified(pandas_df, tmp_path):
    dataset_path = tmp_path / "df.parquet"
    pandas_df.to_parquet(dataset_path)

    with mock.patch("mlflow.pipelines.utils.step.get_pandas_data_profile") as mock_profiling:
        IngestStep.from_pipeline_config(
            pipeline_config={
                "data": {
                    "format": "parquet",
                    "location": str(dataset_path),
                },
                "steps": {"ingest": {"skip_data_profiling": True}},
            },
            pipeline_root=os.getcwd(),
        ).run(output_directory=tmp_path)

    expected_step_card_path = os.path.join(tmp_path, "card.html")
    with open(expected_step_card_path, "r") as f:
        step_card_html_content = f.read()
    assert "Profile of Ingested Dataset" not in step_card_html_content
    mock_profiling.assert_not_called()



import os
import os.path
import re
import shutil

import mlflow
from mlflow import cli
from mlflow.utils import process
from tests.integration.utils import invoke_cli_runner
import pytest
import json
import hashlib

EXAMPLES_DIR = "examples"


def hash_conda_env(conda_env_path):
    # use the same hashing logic as `_get_conda_env_name` in mlflow/utils/conda.py
    return hashlib.sha1(open(conda_env_path).read().encode("utf-8")).hexdigest()


def get_conda_envs():
    stdout = process._exec_cmd(["conda", "env", "list", "--json"]).stdout
    return [os.path.basename(env) for env in json.loads(stdout)["envs"]]


def is_mlflow_conda_env(env_name):
    return re.search(r"^mlflow-\w{40}$", env_name) is not None


def remove_conda_env(env_name):
    process._exec_cmd(["conda", "remove", "--name", env_name, "--yes", "--all"])


def get_free_disk_space():
    # https://stackoverflow.com/a/48929832/6943581
    return shutil.disk_usage("/")[-1] / (2**30)


def is_conda_yaml(path):
    return bool(re.search("conda.ya?ml$", path))


def find_conda_yaml(directory):
    conda_yaml = list(filter(is_conda_yaml, os.listdir(directory)))[0]
    return os.path.join(directory, conda_yaml)


def replace_mlflow_with_dev_version(yml_path):
    with open(yml_path, "r") as f:
        old_src = f.read()
        mlflow_dir = os.path.dirname(mlflow.__path__[0])
        new_src = re.sub(r"- mlflow.*\n", "- {}\n".format(mlflow_dir), old_src)

    with open(yml_path, "w") as f:
        f.write(new_src)


@pytest.fixture(scope="function", autouse=True)
def clean_envs_and_cache():
    yield

    if get_free_disk_space() < 7.0:  # unit: GiB
        process._exec_cmd(["./dev/remove-conda-envs.sh"])


@pytest.fixture(scope="function", autouse=True)
def report_free_disk_space(capsys):
    yield

    with capsys.disabled():
        # pylint: disable-next=print-function
        print(" | Free disk space: {:.1f} GiB".format(get_free_disk_space()), end="")


@pytest.mark.notrackingurimock
@pytest.mark.parametrize(
    "directory, params",
    [
        ("h2o", []),
        ("hyperparam", ["-e", "train", "-P", "epochs=1"]),
        ("hyperparam", ["-e", "random", "-P", "epochs=1"]),
        ("hyperparam", ["-e", "hyperopt", "-P", "epochs=1"]),
        (
            os.path.join("lightgbm", "lightgbm_native"),
            ["-P", "learning_rate=0.1", "-P", "colsample_bytree=0.8", "-P", "subsample=0.9"],
        ),
        (os.path.join("lightgbm", "lightgbm_sklearn"), []),
        ("statsmodels", ["-P", "inverse_method=qr"]),
        ("pytorch", ["-P", "epochs=2"]),
        ("sklearn_logistic_regression", []),
        ("sklearn_elasticnet_wine", ["-P", "alpha=0.5"]),
        (os.path.join("sklearn_elasticnet_diabetes", "linux"), []),
        ("spacy", []),
        (
            os.path.join("xgboost", "xgboost_native"),
            ["-P", "learning_rate=0.3", "-P", "colsample_bytree=0.8", "-P", "subsample=0.9"],
        ),
        (os.path.join("xgboost", "xgboost_sklearn"), []),
        ("fastai", ["-P", "lr=0.02", "-P", "epochs=3"]),
        (os.path.join("pytorch", "MNIST"), ["-P", "max_epochs=1"]),
        (
            os.path.join("pytorch", "BertNewsClassification"),
            ["-P", "max_epochs=1", "-P", "num_samples=100", "-P", "dataset=20newsgroups"],
        ),
        (
            os.path.join("pytorch", "AxHyperOptimizationPTL"),
            ["-P", "max_epochs=10", "-P", "total_trials=1"],
        ),
        (
            os.path.join("pytorch", "IterativePruning"),
            ["-P", "max_epochs=1", "-P", "total_trials=1"],
        ),
        (os.path.join("pytorch", "CaptumExample"), ["-P", "max_epochs=50"]),
        ("supply_chain_security", []),
    ],
)
def test_mlflow_run_example(directory, params, tmpdir):
    example_dir = os.path.join(EXAMPLES_DIR, directory)
    tmp_example_dir = os.path.join(tmpdir.strpath, directory)
    shutil.copytree(example_dir, tmp_example_dir)
    conda_yml_path = find_conda_yaml(tmp_example_dir)
    replace_mlflow_with_dev_version(conda_yml_path)

    # remove old conda environments to free disk space
    envs = list(filter(is_mlflow_conda_env, get_conda_envs()))
    current_env_name = "mlflow-" + hash_conda_env(conda_yml_path)
    envs_to_remove = list(filter(lambda e: e != current_env_name, envs))
    for env in envs_to_remove:
        remove_conda_env(env)

    cli_run_list = [tmp_example_dir] + params
    invoke_cli_runner(cli.run, cli_run_list)


@pytest.mark.notrackingurimock
@pytest.mark.parametrize(
    "directory, command",
    [
        ("docker", ["docker", "build", "-t", "mlflow-docker-example", "-f", "Dockerfile", "."]),
        ("gluon", ["python", "train.py"]),
        ("keras", ["python", "train.py"]),
        (
            os.path.join("lightgbm", "lightgbm_native"),
            [
                "python",
                "train.py",
                "--learning-rate",
                "0.2",
                "--colsample-bytree",
                "0.8",
                "--subsample",
                "0.9",
            ],
        ),
        (os.path.join("lightgbm", "lightgbm_sklearn"), ["python", "train.py"]),
        ("statsmodels", ["python", "train.py", "--inverse-method", "qr"]),
        ("quickstart", ["python", "mlflow_tracking.py"]),
        ("remote_store", ["python", "remote_server.py"]),
        (
            os.path.join("xgboost", "xgboost_native"),
            [
                "python",
                "train.py",
                "--learning-rate",
                "0.2",
                "--colsample-bytree",
                "0.8",
                "--subsample",
                "0.9",
            ],
        ),
        (os.path.join("xgboost", "xgboost_sklearn"), ["python", "train.py"]),
        ("catboost", ["python", "train.py"]),
        ("prophet", ["python", "train.py"]),
        ("sklearn_autolog", ["python", "linear_regression.py"]),
        ("sklearn_autolog", ["python", "pipeline.py"]),
        ("sklearn_autolog", ["python", "grid_search_cv.py"]),
        ("pyspark_ml_autologging", ["python", "logistic_regression.py"]),
        ("pyspark_ml_autologging", ["python", "one_vs_rest.py"]),
        ("pyspark_ml_autologging", ["python", "pipeline.py"]),
        ("shap", ["python", "regression.py"]),
        ("shap", ["python", "binary_classification.py"]),
        ("shap", ["python", "multiclass_classification.py"]),
        ("shap", ["python", "explainer_logging.py"]),
        ("ray_serve", ["python", "train_model.py"]),
        ("pip_requirements", ["python", "pip_requirements.py"]),
        ("fastai", ["python", "train.py", "--lr", "0.02", "--epochs", "3"]),
        ("pmdarima", ["python", "train.py"]),
        ("evaluation", ["python", "evaluate_on_binary_classifier.py"]),
        ("evaluation", ["python", "evaluate_on_multiclass_classifier.py"]),
        ("evaluation", ["python", "evaluate_on_regressor.py"]),
        ("evaluation", ["python", "evaluate_with_custom_metrics.py"]),
        ("evaluation", ["python", "evaluate_with_custom_metrics_comprehensive.py"]),
        ("evaluation", ["python", "evaluate_with_model_validation.py"]),
        ("diviner", ["python", "train.py"]),
        ("spark_udf", ["python", "spark_udf_datetime.py"]),
        ("pyfunc", ["python", "train.py"]),
    ],
)
def test_command_example(directory, command):
    cwd_dir = os.path.join(EXAMPLES_DIR, directory)
    process._exec_cmd(command, cwd=cwd_dir)



import uuid

import pytest

from mlflow.exceptions import MlflowException
from mlflow.pipelines.cards import BaseCard, CardTab


class ProfileReport:
    def to_html(self):
        return "pandas-profiling"


def test_verify_card_information():
    card = BaseCard(
        pipeline_name="fake pipeline",
        step_name="fake step",
    )
    (
        card.add_tab(
            "First tab",
            """
        <h3 class="section-title">Markdown:</h3>
        {{ MARKDOWN_1 }}<hr>
        <h3 class="section-title">Other:</h3>
        {{ HTML_1 }}""",
        )
        .add_markdown("MARKDOWN_1", "#### Hello, world!")
        .add_html("HTML_1", "<span style='color:blue'>blue</span>")
    )
    card.add_tab("Profile 1", "{{PROFILE}}").add_pandas_profile("PROFILE", ProfileReport())
    card.add_tab("Profile 2", "{{PROFILE}}").add_pandas_profile("PROFILE", ProfileReport())
    card.add_text("1,2,3.")

    expected_html = """
        <h3 class="section-title">Markdown:</h3>
        <h4>Hello, world!</h4><hr>
        <h3 class="section-title">Other:</h3>
        <span style='color:blue'>blue</span></div>
    """
    assert expected_html in card.to_html()
    assert card.to_text() == "1,2,3."
    assert all(card.get_tab(name) is not None for name in ["Profile 1", "Profile 2", "First tab"])
    assert all(card.get_tab(name) is None for name in ["", "!x", "fake tab 3"])


def test_card_tab_works():
    tab = (
        CardTab("tab", "{{MARKDOWN_1}}{{HTML_1}}{{PROFILE_1}}")
        .add_html("HTML_1", "<span style='color:blue'>blue</span>")
        .add_markdown("MARKDOWN_1", "#### Hello, world!")
        .add_pandas_profile("PROFILE_1", ProfileReport())
    )
    assert (
        tab.to_html()
        == "<h4>Hello, world!</h4><span style='color:blue'>blue</span>"
        + "<iframe srcdoc='pandas-profiling' width='100%' height='500' frameborder='0'></iframe>"
    )


def test_card_tab_fails_with_invalid_variable():
    with pytest.raises(MlflowException, match=r"(not a valid template variable)"):
        CardTab("tab", "{{MARKDOWN_1}}").add_html("HTML_1", "<span style='color:blue'>blue</span>")


def test_render_table():
    import pandas as pd
    import numpy as np

    assert "</table>" in BaseCard.render_table(pd.DataFrame({"a": [1, 2], "b": [3, 4]}))
    assert "</table>" in BaseCard.render_table(pd.DataFrame({"a": [1, 2], "b": [3, 4]}).style)
    assert "</table>" in BaseCard.render_table([(1, 2), (3, 4)], columns=["a", "b"])
    assert "</table>" in BaseCard.render_table([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    assert "</table>" in BaseCard.render_table({"a": [1, 2], "b": [3, 4]})
    assert "</table>" in BaseCard.render_table(np.array([[1, 2], [3, 4]]), columns=["a", "b"])

    col_name = uuid.uuid4().hex
    assert col_name in BaseCard.render_table({"a": [1, 2], "b": [3, 4], col_name: [5, 6]})
    assert col_name not in BaseCard.render_table(
        {"a": [1, 2], "b": [3, 4], col_name: [5, 6]}, columns=["a", "b"]
    )



import os
import pytest
import shutil
from pathlib import Path

import mlflow
from mlflow.utils.file_utils import read_yaml
from mlflow.pipelines.utils import _PIPELINE_CONFIG_FILE_NAME
from mlflow.pipelines.steps.evaluate import EvaluateStep
from mlflow.pipelines.steps.register import RegisterStep
from mlflow.store.tracking.sqlalchemy_store import SqlAlchemyStore

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    clear_custom_metrics_module_cache,
    setup_model_and_evaluate,
    tmp_pipeline_exec_path,
    tmp_pipeline_root_path,
)  # pylint: enable=unused-import


@pytest.fixture
def registry_uri_path(tmp_path) -> Path:
    previousRegistryUri = ""
    try:
        previousRegistryUri = mlflow.get_registry_uri()
        path = tmp_path.joinpath("registry.db")
        db_url = "sqlite:///%s" % path
        SqlAlchemyStore(db_url, "register_model")
        yield db_url
    finally:
        os.remove(path)
        shutil.rmtree("register_model")
        mlflow.set_registry_uri(previousRegistryUri)


@pytest.mark.usefixtures("clear_custom_metrics_module_cache")
@pytest.mark.parametrize(
    "mae_threshold,register_flag",
    [
        (-1, ""),
        (1_000_000, ""),
        (-1, "allow_non_validated_model: true"),
        (1_000_000, "allow_non_validated_model: true"),
    ],
)
def test_register_step_run(
    tmp_pipeline_root_path: Path,
    tmp_pipeline_exec_path: Path,
    mae_threshold: int,
    register_flag: str,
):
    evaluate_step_output_dir, register_step_output_dir = setup_model_and_evaluate(
        tmp_pipeline_exec_path
    )
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 1_000_000
      - metric: mean_absolute_error
        threshold: {mae_threshold}
      - metric: weighted_mean_squared_error
        threshold: 1_000_000
  register:
    model_name: "demo_model"
    {allow_non_validated_model}
metrics:
  custom:
    - name: weighted_mean_squared_error
      function: weighted_mean_squared_error
      greater_is_better: False
""".format(
            tracking_uri=mlflow.get_tracking_uri(),
            mae_threshold=mae_threshold,
            allow_non_validated_model=register_flag,
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_steps_dir.joinpath("custom_metrics.py").write_text(
        """
def weighted_mean_squared_error(eval_df, builtin_metrics):
    from sklearn.metrics import mean_squared_error

    return {
        "weighted_mean_squared_error": mean_squared_error(
            eval_df["prediction"],
            eval_df["target"],
            sample_weight=1 / eval_df["prediction"].values,
        )
    }
"""
    )
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))
    assert len(mlflow.tracking.MlflowClient().list_registered_models()) == 0
    register_step = RegisterStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    register_step._run(str(register_step_output_dir))
    model_validation_status_path = evaluate_step_output_dir.joinpath("model_validation_status")
    assert model_validation_status_path.exists()
    expected_status = "REJECTED" if mae_threshold < 0 else "VALIDATED"
    assert model_validation_status_path.read_text() == expected_status
    assert len(mlflow.tracking.MlflowClient().list_registered_models()) == (
        1 if expected_status == "VALIDATED" else 0
    )


@pytest.mark.usefixtures("clear_custom_metrics_module_cache")
@pytest.mark.parametrize("register_flag", ["", "allow_non_validated_model: true"])
def test_register_with_no_validation_criteria(
    tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path, register_flag: str
):
    evaluate_step_output_dir, register_step_output_dir = setup_model_and_evaluate(
        tmp_pipeline_exec_path
    )
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
  register:
    model_name: "demo_model"
    {allow_non_validated_model}
""".format(
            tracking_uri=mlflow.get_tracking_uri(),
            allow_non_validated_model=register_flag,
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))
    assert len(mlflow.tracking.MlflowClient().list_registered_models()) == 0
    register_step = RegisterStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    register_step._run(str(register_step_output_dir))
    model_validation_status_path = evaluate_step_output_dir.joinpath("model_validation_status")
    assert model_validation_status_path.exists()
    assert model_validation_status_path.read_text() == "UNKNOWN"
    assert len(mlflow.tracking.MlflowClient().list_registered_models()) == (
        0 if register_flag == "" else 1
    )


def test_usage_tracking_correctly_added(
    tmp_pipeline_root_path: Path,
    tmp_pipeline_exec_path: Path,
):
    evaluate_step_output_dir, register_step_output_dir = setup_model_and_evaluate(
        tmp_pipeline_exec_path
    )
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 1_000_000
      - metric: mean_absolute_error
        threshold: 1_000_000
      - metric: weighted_mean_squared_error
        threshold: 1_000_000
  register:
    model_name: "demo_model"
metrics:
  custom:
    - name: weighted_mean_squared_error
      function: weighted_mean_squared_error
      greater_is_better: False
""".format(
            tracking_uri=mlflow.get_tracking_uri(),
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_steps_dir.joinpath("custom_metrics.py").write_text(
        """
def weighted_mean_squared_error(eval_df, builtin_metrics):
    from sklearn.metrics import mean_squared_error

    return {
        "weighted_mean_squared_error": mean_squared_error(
            eval_df["prediction"],
            eval_df["target"],
            sample_weight=1 / eval_df["prediction"].values,
        )
    }
"""
    )
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))
    register_step = RegisterStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    register_step._run(str(register_step_output_dir))
    registered_models = mlflow.tracking.MlflowClient().list_registered_models()
    latest_tag = registered_models[0].latest_versions[0].tags
    assert latest_tag["mlflow.source.type"] == "PIPELINE"
    assert latest_tag["mlflow.pipeline.template.name"] == "regression/v1"


def test_register_uri(
    tmp_pipeline_root_path: Path,
    tmp_pipeline_exec_path: Path,
    registry_uri_path: Path,
):
    evaluate_step_output_dir, register_step_output_dir = setup_model_and_evaluate(
        tmp_pipeline_exec_path
    )
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    registry_uri = registry_uri_path
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
model_registry:
  uri: {registry_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 1_000_000
      - metric: mean_absolute_error
        threshold: 1_000_000
      - metric: weighted_mean_squared_error
        threshold: 1_000_000
  register:
    model_name: "demo_model"
metrics:
  custom:
    - name: weighted_mean_squared_error
      function: weighted_mean_squared_error
      greater_is_better: False
""".format(
            tracking_uri=mlflow.get_tracking_uri(),
            registry_uri=registry_uri,
        )
    )
    pipeline_steps_dir = tmp_pipeline_root_path.joinpath("steps")
    pipeline_steps_dir.mkdir(parents=True)
    pipeline_steps_dir.joinpath("custom_metrics.py").write_text(
        """
def weighted_mean_squared_error(eval_df, builtin_metrics):
    from sklearn.metrics import mean_squared_error

    return {
        "weighted_mean_squared_error": mean_squared_error(
            eval_df["prediction"],
            eval_df["target"],
            sample_weight=1 / eval_df["prediction"].values,
        )
    }
"""
    )
    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))
    register_step = RegisterStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    register_step._run(str(register_step_output_dir))
    assert mlflow.get_registry_uri() == registry_uri


def test_register_step_writes_card_with_model_link_and_version_link_on_databricks(
    monkeypatch, tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path
):
    workspace_host = "https://dev.databricks.com"
    workspace_id = 123456
    workspace_url = f"{workspace_host}?o={workspace_id}"

    monkeypatch.setenv("_DATABRICKS_WORKSPACE_HOST", workspace_host)
    monkeypatch.setenv("_DATABRICKS_WORKSPACE_ID", workspace_id)

    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "regression/v1"
target_col: "y"
experiment:
  tracking_uri: {tracking_uri}
steps:
  evaluate:
    validation_criteria:
      - metric: root_mean_squared_error
        threshold: 1_000_000
  register:
    model_name: "demo_model"
""".format(
            tracking_uri=mlflow.get_tracking_uri()
        )
    )

    evaluate_step_output_dir, register_step_output_dir = setup_model_and_evaluate(
        tmp_pipeline_exec_path
    )

    pipeline_config = read_yaml(tmp_pipeline_root_path, _PIPELINE_CONFIG_FILE_NAME)
    evaluate_step = EvaluateStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    evaluate_step._run(str(evaluate_step_output_dir))

    register_step = RegisterStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    register_step._run(str(register_step_output_dir))

    train_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "train", "outputs")
    with open(train_step_output_dir / "run_id") as f:
        run_id = f.read()

    assert (register_step_output_dir / "card.html").exists()
    with open(register_step_output_dir / "card.html", "r") as f:
        step_card_content = f.read()

    assert f"<a href={workspace_url}#mlflow/models/demo_model/1" in step_card_content
    assert (
        f"<a href={workspace_url}#mlflow/experiments/1/runs/{run_id}/artifactPath/train/model>"
        in step_card_content
    )



import mlflow.diviner

get_default_pip_requirements = mlflow.diviner.get_default_pip_requirements
get_default_conda_env = mlflow.diviner.get_default_conda_env
save_model = mlflow.diviner.save_model
log_model = mlflow.diviner.log_model
_load_model = mlflow.diviner._load_model
_get_diviner_instance_type = mlflow.diviner._get_diviner_instance_type
_DivinerModelWrapper = mlflow.diviner._DivinerModelWrapper
_load_pyfunc = mlflow.diviner._load_pyfunc
load_model = mlflow.diviner.load_model



import mlflow.keras

get_default_pip_requirements = mlflow.keras.get_default_pip_requirements
get_default_conda_env = mlflow.keras.get_default_conda_env
save_model = mlflow.keras.save_model
log_model = mlflow.keras.log_model
_load_model = mlflow.keras._load_model
_KerasModelWrapper = mlflow.keras._KerasModelWrapper
_load_pyfunc = mlflow.keras._load_pyfunc
load_model = mlflow.keras.load_model
autolog = mlflow.keras.autolog
_save_custom_objects = mlflow.keras._save_custom_objects


import os
import requests
import time
import webbrowser

class Auth0():
    auth0url = "https://dev-pg1h84k1.us.auth0.com"
    url = auth0url + '/oauth/device/code'
    userURL = auth0url + '/userinfo'
    clientId = "qhgH8CCF8riL9XYOZr8EPym8lxq3XEd3"
    audience = "https://grpc-api-gateway-d8q71ttn.uc.gateway.dev/"
    clientCredentials = {"client_id": clientId , "scope" : "openid email profile newron-server" , "audience" : audience}

    def __init__(self) -> None:
        pass

    def authenticate(self):
        x = requests.post(self.url, json = self.clientCredentials)
        if x.status_code != 200:
            print("Error in verification: " + str(x.status_code))
            print(x.text)
            exit()
        resp = x.json()
        print("\n\n************************************************************\n\n")
        print("Please verify the following code on the device: " + str(resp['user_code']))
        print("\n\n************************************************************\n\n")
        print("Please visit the following link " + str(resp['verification_uri_complete']))
        print("\n\n************************************************************\n\n")
        # Step 2: Open Webbrowser
        webbrowser.open(resp["verification_uri_complete"])
        # Step 3: Poll for token
        pollUrl = self.auth0url + "/oauth/token"
        pollObj = { 
            "client_id": self.clientId,
            "device_code": resp["device_code"],
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
        }
        maxPolls = 60
        op = requests.post(pollUrl, json = pollObj)
        opJson = op.json()
        failed_polls = 0
        # print(op.text)
        # print(op.status_code)
        while op.status_code != 200 and maxPolls > 0:
            op = requests.post(pollUrl, json = pollObj)
            opJson = op.json()
            maxPolls -= 1
            if op.status_code == 200:
                print("Authorization Successful")
                break;
            if op.status_code == 400 or op.status_code == 401:
                print("Authorization Failed")
                exit();
            if opJson["error"] == "invalid_grant":
                print("Authorization Failed")
                break;
            if opJson["error"] == "authorization_pending":
                print("Authorization Pending")
                time.sleep(2)
            elif opJson["error"] == "slow_down":
                if failed_polls == 0:
                    time.sleep(10)
                else:
                    print("Waiting for Authorization")
                    time.sleep(8)
                failed_polls+=1
            else:
                time.sleep(10)
                break
        if(op.status_code != 200 or maxPolls == 0):
            # print("Error: " + op.text)
            exit()
        token = op.json()["access_token"]
        headers = {"Authorization": "Bearer " + token}
        userResponse = requests.get(self.userURL, headers = headers)
        if not userResponse.json()["email_verified"]:
            print("Pease Verify your email")
        userResponse = userResponse.json()
        userResponse["access_token"] = token
        return userResponse


from  mlflow.pyspark.ml import (_read_log_model_allowlist_from_file,
                                _read_log_model_allowlist_from_file,
                                _get_warning_msg_for_skip_log_model,
                                _should_log_model,
                                _get_estimator_info_tags,
                                _is_parameter_search_estimator,
                                _is_parameter_search_model,
                                _should_log_hierarchy,
                                _traverse_stage,
                                _get_uid_to_indexed_name_map,
                                _gen_stage_hierarchy_recursively,
                                _gen_estimator_metadata,
                                _get_param_map,
                                _get_stage_type_params,
                                _get_instance_param_map_recursively,
                                _get_instance_param_map,
                                _create_child_runs_for_parameter_search,
                                _log_parameter_search_results_as_artifact,
                                _get_warning_msg_for_fit_call_with_a_list_of_params,
                                _get_tuning_param_maps,
                                _get_param_search_metrics_and_best_index,
                                _log_estimator_params,
                                _AutologgingMetricsManager,
                                _get_columns_with_unsupported_data_type,
                                autolog,
                                _log_pretraining_metadata,
                                _log_posttraining_metadata,
                                fit_mlflow,
                                patched_fit,
                                patched_transform,
                                patched_evaluate,
                                )


import os
import pathlib
import shutil
import time
from unittest import mock

import pandas as pd
import pytest

from mlflow.pipelines import Pipeline
from mlflow.pipelines.steps.ingest import IngestStep
from mlflow.pipelines.steps.split import SplitStep
from mlflow.pipelines.steps.transform import TransformStep
from mlflow.pipelines.step import StepStatus
from mlflow.pipelines.utils.execution import (
    _get_or_create_execution_directory,
    run_pipeline_step,
    get_step_output_path,
)

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    BaseStepImplemented,
    enter_pipeline_example_directory,
    enter_test_pipeline_directory,
)


@pytest.fixture
def pandas_df():
    df = pd.DataFrame.from_dict(
        {
            "A": ["x", "y", "z"],
            "B": [1, 2, 3],
            "C": [-9.2, 82.5, 3.40],
        }
    )
    df.index.rename("index", inplace=True)
    return df


@pytest.fixture
def test_pipeline(
    enter_test_pipeline_directory, pandas_df, tmp_path
):  # pylint: disable=unused-argument
    dataset_path = tmp_path / "df.parquet"
    pandas_df.to_parquet(dataset_path)
    ingest_step = IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": str(dataset_path),
            }
        },
        pipeline_root=os.getcwd(),
    )
    split_step = SplitStep.from_pipeline_config(
        pipeline_config={
            "target_col": "C",
            "steps": {
                "split": {},
            },
        },
        pipeline_root=os.getcwd(),
    )
    with open(pathlib.Path().cwd() / "steps" / "transform.py", "w") as f:
        f.write(
            "\n".join(
                [
                    "from sklearn.pipeline import Pipeline",
                    "from sklearn.preprocessing import FunctionTransformer",
                    "",
                    "def add_column(df):",
                    "    df['useless'] = 'useless'",
                    "    return df",
                    "",
                    "def transform_fn():",
                    "    return Pipeline(steps=[('add_column', FunctionTransformer(add_column))])",
                ]
            )
        )
    transform_step = TransformStep.from_pipeline_config(
        pipeline_config={
            "target_col": "C",
            "steps": {
                "transform": {
                    "transformer_method": "steps.transform.transform_fn",
                },
            },
        },
        pipeline_root=os.getcwd(),
    )
    return [ingest_step, split_step, transform_step]


@pytest.fixture(autouse=True)
def clean_test_pipeline(enter_test_pipeline_directory):  # pylint: disable=unused-argument
    Pipeline(profile="local").clean()
    try:
        yield
    finally:
        Pipeline(profile="local").clean()


def test_get_or_create_execution_directory_is_idempotent(tmp_path):
    class TestStep(BaseStepImplemented):
        def __init__(self):  # pylint: disable=super-init-not-called
            pass

        @property
        def name(self):
            return "test_step"

    test_step = TestStep()

    def assert_expected_execution_directory_contents_exist(execution_dir_path):
        assert (execution_dir_path / "Makefile").exists()
        assert (execution_dir_path / "steps").exists()
        assert (execution_dir_path / "steps" / test_step.name / "outputs").exists()

    execution_dir_path_1 = pathlib.Path(
        _get_or_create_execution_directory(
            pipeline_root_path=tmp_path, pipeline_steps=[test_step], template="regression/v1"
        )
    )
    execution_dir_path_2 = pathlib.Path(
        _get_or_create_execution_directory(
            pipeline_root_path=tmp_path, pipeline_steps=[test_step], template="regression/v1"
        )
    )
    assert execution_dir_path_1 == execution_dir_path_2
    assert_expected_execution_directory_contents_exist(execution_dir_path_1)

    shutil.rmtree(execution_dir_path_1)

    # Simulate a failure with Makefile creation
    with mock.patch(
        "mlflow.pipelines.utils.execution._create_makefile",
        side_effect=Exception("Makefile creation failed"),
    ), pytest.raises(Exception, match="Makefile creation failed"):
        _get_or_create_execution_directory(
            pipeline_root_path=tmp_path, pipeline_steps=[test_step], template="regression/v1"
        )

    # Verify that the directory exists but is empty due to short circuiting after
    # failed Makefile creation
    assert execution_dir_path_1.exists()
    assert next(execution_dir_path_1.iterdir(), None) == None

    # Re-create the execution directory and verify that all expected contents are present
    execution_dir_path_3 = pathlib.Path(
        _get_or_create_execution_directory(
            pipeline_root_path=tmp_path, pipeline_steps=[test_step], template="regression/v1"
        )
    )
    assert execution_dir_path_3 == execution_dir_path_1
    assert_expected_execution_directory_contents_exist(execution_dir_path_3)

    shutil.rmtree(execution_dir_path_1)

    # Simulate a failure with step-specific directory creation
    with mock.patch(
        "mlflow.pipelines.utils.execution._get_step_output_directory_path",
        side_effect=Exception("Step directory creation failed"),
    ), pytest.raises(Exception, match="Step directory creation failed"):
        _get_or_create_execution_directory(
            pipeline_root_path=tmp_path, pipeline_steps=[test_step], template="regression/v1"
        )

    # Verify that the directory exists & that a Makefile is present but step-specific directories
    # were not created due to failures
    assert execution_dir_path_1.exists()
    assert [path.name for path in execution_dir_path_1.iterdir()] == ["Makefile"]

    # Re-create the execution directory and verify that all expected contents are present
    execution_dir_path_4 = pathlib.Path(
        _get_or_create_execution_directory(
            pipeline_root_path=tmp_path, pipeline_steps=[test_step], template="regression/v1"
        )
    )
    assert execution_dir_path_4 == execution_dir_path_1
    assert_expected_execution_directory_contents_exist(execution_dir_path_4)


def test_run_pipeline_step_sets_environment_as_expected(tmp_path):
    class TestStep1(BaseStepImplemented):
        def __init__(self):  # pylint: disable=super-init-not-called
            self.step_config = {}

        @property
        def name(self):
            return "test_step_1"

        @property
        def environment(self):
            return {"A": "B"}

    class TestStep2(BaseStepImplemented):
        def __init__(self):  # pylint: disable=super-init-not-called
            self.step_config = {}

        @property
        def name(self):
            return "test_step_2"

        @property
        def environment(self):
            return {"C": "D"}

    with mock.patch("mlflow.pipelines.utils.execution._exec_cmd") as mock_run_in_subprocess:
        pipeline_steps = [TestStep1(), TestStep2()]
        run_pipeline_step(
            pipeline_root_path=tmp_path,
            pipeline_steps=pipeline_steps,
            target_step=pipeline_steps[0],
            template="regression/v1",
        )

    _, subprocess_call_kwargs = mock_run_in_subprocess.call_args
    assert subprocess_call_kwargs.get("extra_env") == {"A": "B", "C": "D"}


def run_test_pipeline_step(pipeline_steps, target_step):
    return run_pipeline_step(
        pipeline_root_path=os.getcwd(),
        pipeline_steps=pipeline_steps,
        target_step=target_step,
        template="regression/v1",
    )


def get_test_pipeline_step_output_directory(step):
    return get_step_output_path(os.getcwd(), step.name, "")


def get_test_pipeline_step_execution_state(step):
    return step.get_execution_state(get_test_pipeline_step_output_directory(step))


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_run_pipeline_step_maintains_execution_status_correctly(pandas_df, tmp_path):
    dataset_path = tmp_path / "df.parquet"
    pandas_df.to_parquet(dataset_path)

    ingest_step_good = IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": str(dataset_path),
            }
        },
        pipeline_root=os.getcwd(),
    )

    assert get_test_pipeline_step_execution_state(ingest_step_good).status == StepStatus.UNKNOWN
    assert get_test_pipeline_step_execution_state(ingest_step_good).last_updated_timestamp == 0
    curr_time = time.time()
    run_test_pipeline_step([ingest_step_good], ingest_step_good)
    assert get_test_pipeline_step_execution_state(ingest_step_good).status == StepStatus.SUCCEEDED
    assert (
        get_test_pipeline_step_execution_state(ingest_step_good).last_updated_timestamp >= curr_time
    )

    ingest_step_bad = IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": "badlocation",
            }
        },
        pipeline_root=os.getcwd(),
    )
    curr_time = time.time()
    run_test_pipeline_step([ingest_step_bad], ingest_step_bad)
    assert get_test_pipeline_step_execution_state(ingest_step_bad).status == StepStatus.FAILED
    assert (
        get_test_pipeline_step_execution_state(ingest_step_bad).last_updated_timestamp >= curr_time
    )


def test_run_pipeline_step_returns_expected_result(test_pipeline):
    ingest_step, split_step, _ = test_pipeline

    assert run_test_pipeline_step(test_pipeline, ingest_step) == ingest_step
    assert run_test_pipeline_step(test_pipeline, split_step) == split_step
    assert run_test_pipeline_step(test_pipeline, ingest_step) == ingest_step

    ingest_step_bad = IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": "badlocation",
            }
        },
        pipeline_root=os.getcwd(),
    )

    assert run_test_pipeline_step([ingest_step_bad, split_step], ingest_step_bad) == ingest_step_bad
    assert run_test_pipeline_step([ingest_step_bad, split_step], split_step) == ingest_step_bad


def test_run_pipeline_with_ingest_step_as_target_never_caches(test_pipeline):
    ingest_step, _, _ = test_pipeline

    def get_step_outputs_with_timestamps(step):
        output_directory = get_test_pipeline_step_output_directory(step)
        return {
            path: os.path.getmtime(os.path.join(output_directory, path))
            for path in os.listdir(output_directory)
        }

    curr_time = time.time()
    run_test_pipeline_step(test_pipeline, ingest_step)
    step_outputs_with_timestamps_1 = get_step_outputs_with_timestamps(ingest_step)
    assert step_outputs_with_timestamps_1
    assert get_test_pipeline_step_execution_state(ingest_step).last_updated_timestamp >= curr_time

    curr_time = time.time()
    run_test_pipeline_step(test_pipeline, ingest_step)
    step_outputs_with_timestamps_2 = get_step_outputs_with_timestamps(ingest_step)
    assert step_outputs_with_timestamps_2
    assert get_test_pipeline_step_execution_state(ingest_step).last_updated_timestamp >= curr_time

    assert step_outputs_with_timestamps_2 != step_outputs_with_timestamps_1


@pytest.mark.parametrize("target_step", ["split", "transform"])
def test_run_pipeline_step_caches(test_pipeline, target_step):
    _, split_step, transform_step = test_pipeline
    target_step = split_step if target_step == "split" else transform_step

    def get_step_outputs_with_timestamps(step):
        output_directory = get_test_pipeline_step_output_directory(step)
        return {
            path: os.path.getmtime(os.path.join(output_directory, path))
            for path in os.listdir(output_directory)
        }

    curr_time = time.time()
    run_test_pipeline_step(test_pipeline, target_step)
    step_outputs_with_timestamps_1 = get_step_outputs_with_timestamps(target_step)
    assert step_outputs_with_timestamps_1
    step_execution_state_1 = get_test_pipeline_step_execution_state(target_step)
    assert step_execution_state_1.status == StepStatus.SUCCEEDED
    assert step_execution_state_1.last_updated_timestamp >= curr_time

    curr_time = time.time()
    run_test_pipeline_step(test_pipeline, target_step)
    step_outputs_with_timestamps_2 = get_step_outputs_with_timestamps(target_step)
    assert step_outputs_with_timestamps_2
    step_execution_state_2 = get_test_pipeline_step_execution_state(target_step)
    assert (
        step_execution_state_2.last_updated_timestamp
        == step_execution_state_1.last_updated_timestamp
    )

    assert step_outputs_with_timestamps_2 == step_outputs_with_timestamps_1


def test_run_pipeline_with_ingest_step_as_target_clears_downstream_step_state(test_pipeline):
    ingest_step, split_step, transform_step = test_pipeline

    curr_time = time.time()
    run_test_pipeline_step(test_pipeline, transform_step)
    for step in test_pipeline:
        assert get_test_pipeline_step_execution_state(step).status == StepStatus.SUCCEEDED
        assert get_test_pipeline_step_execution_state(step).last_updated_timestamp >= curr_time
        assert os.listdir(get_test_pipeline_step_output_directory(step))

    curr_time = time.time()
    run_test_pipeline_step(test_pipeline, ingest_step)
    assert get_test_pipeline_step_execution_state(ingest_step).status == StepStatus.SUCCEEDED
    assert get_test_pipeline_step_execution_state(ingest_step).last_updated_timestamp >= curr_time
    assert os.listdir(get_test_pipeline_step_output_directory(ingest_step))
    for step in [split_step, transform_step]:
        assert get_test_pipeline_step_execution_state(step).status == StepStatus.UNKNOWN
        assert get_test_pipeline_step_execution_state(step).last_updated_timestamp == 0
        assert not os.listdir(get_test_pipeline_step_output_directory(step))


def test_run_pipeline_step_after_change_clears_downstream_step_state(test_pipeline):
    ingest_step, _, transform_step = test_pipeline
    curr_time = time.time()

    run_test_pipeline_step(test_pipeline, transform_step)
    for step in test_pipeline:
        assert get_test_pipeline_step_execution_state(step).status == StepStatus.SUCCEEDED
        assert get_test_pipeline_step_execution_state(step).last_updated_timestamp >= curr_time
        assert os.listdir(get_test_pipeline_step_output_directory(step))

    updated_split_step = SplitStep.from_pipeline_config(
        pipeline_config={
            "target_col": "C",
            "steps": {
                "split": {
                    "split_ratios": [0.8, 0.1, 0.1],
                },
            },
        },
        pipeline_root=os.getcwd(),
    )
    run_test_pipeline_step([ingest_step, updated_split_step, transform_step], updated_split_step)
    for step in [ingest_step, updated_split_step]:
        assert get_test_pipeline_step_execution_state(step).status == StepStatus.SUCCEEDED
        assert get_test_pipeline_step_execution_state(step).last_updated_timestamp >= curr_time
        assert os.listdir(get_test_pipeline_step_output_directory(step))

    assert get_test_pipeline_step_execution_state(transform_step).status == StepStatus.UNKNOWN
    assert get_test_pipeline_step_execution_state(transform_step).last_updated_timestamp == 0
    assert not os.listdir(get_test_pipeline_step_output_directory(transform_step))


def test_run_pipeline_step_without_change_preserves_state_of_all_pipeline_steps(test_pipeline):
    _, split_step, transform_step = test_pipeline
    curr_time = time.time()

    run_test_pipeline_step(test_pipeline, transform_step)
    step_to_execution_state = {}
    for step in test_pipeline:
        step_execution_state = get_test_pipeline_step_execution_state(step)
        step_to_execution_state[step] = step_execution_state
        assert step_execution_state.status == StepStatus.SUCCEEDED
        assert step_execution_state.last_updated_timestamp >= curr_time

    run_test_pipeline_step(test_pipeline, split_step)
    for step in test_pipeline:
        step_execution_state = get_test_pipeline_step_execution_state(step)
        prev_execution_state = step_to_execution_state[step]
        assert step_execution_state.status == prev_execution_state.status
        assert (
            step_execution_state.last_updated_timestamp
            == prev_execution_state.last_updated_timestamp
        )


def test_run_pipeline_step_failure_clears_downstream_step_state(test_pipeline):
    ingest_step, split_step, _ = test_pipeline

    curr_time = time.time()
    run_test_pipeline_step(test_pipeline, split_step)
    for step in [ingest_step, split_step]:
        assert get_test_pipeline_step_execution_state(step).status == StepStatus.SUCCEEDED
        assert get_test_pipeline_step_execution_state(step).last_updated_timestamp >= curr_time
        assert os.listdir(get_test_pipeline_step_output_directory(step))

    ingest_step_bad = IngestStep.from_pipeline_config(
        pipeline_config={
            "data": {
                "format": "parquet",
                "location": "badlocation",
            }
        },
        pipeline_root=os.getcwd(),
    )

    curr_time = time.time()
    run_test_pipeline_step([ingest_step_bad, split_step], ingest_step_bad)
    assert get_test_pipeline_step_execution_state(ingest_step_bad).status == StepStatus.FAILED
    assert (
        get_test_pipeline_step_execution_state(ingest_step_bad).last_updated_timestamp >= curr_time
    )
    assert get_test_pipeline_step_execution_state(split_step).status == StepStatus.UNKNOWN
    assert get_test_pipeline_step_execution_state(split_step).last_updated_timestamp == 0
    assert not os.listdir(get_test_pipeline_step_output_directory(split_step))



import os
import json
import functools
import pickle
import pytest
import yaml
import numpy as np
import pandas as pd
import lightgbm as lgb
import matplotlib as mpl
from packaging.version import Version
from sklearn import datasets
from unittest import mock

import newron
import newron.lightgbm
from newron.lightgbm import _autolog_callback
from newron.models import Model
from mlflow.models.utils import _read_example
from newron import NewronClient
from mlflow.utils.autologging_utils import picklable_exception_safe_function, BatchMetricsLogger
from unittest.mock import patch

mpl.use("Agg")


def get_latest_run():
    client = NewronClient()
    return client.get_run(client.list_run_infos(experiment_id="0")[0].run_id)


def get_model_conf(artifact_uri, model_subpath="model"):
    model_conf_path = os.path.join(artifact_uri, model_subpath, "MLmodel")
    return Model.load(model_conf_path)


@pytest.fixture(scope="module")
def bst_params():
    return {
        "objective": "multiclass",
        "num_class": 3,
    }


@pytest.fixture(scope="module")
def train_set():
    iris = datasets.load_iris()
    X = pd.DataFrame(iris.data[:, :2], columns=iris.feature_names[:2])
    y = iris.target
    # set free_raw_data False to use raw data later.
    return lgb.Dataset(X, y, free_raw_data=False)


def test_lgb_autolog_ends_auto_created_run(bst_params, train_set):
    newron.lightgbm.autolog()
    lgb.train(bst_params, train_set, num_boost_round=1)
    assert newron.active_run() is None


def test_lgb_autolog_persists_manually_created_run(bst_params, train_set):
    newron.lightgbm.autolog()
    with newron.start_run() as run:
        lgb.train(bst_params, train_set, num_boost_round=1)
        assert newron.active_run()
        assert newron.active_run().info.run_id == run.info.run_id


def test_lgb_autolog_logs_default_params(bst_params, train_set):
    newron.lightgbm.autolog()
    lgb.train(bst_params, train_set)
    run = get_latest_run()
    params = run.data.params

    expected_params = {
        "num_boost_round": 100,
        "feature_name": "auto",
        "categorical_feature": "auto",
        "keep_training_booster": False,
    }
    if Version(lgb.__version__) <= Version("3.3.1"):
        # The parameter `verbose_eval` in `lightgbm.train` is removed in this PR:
        # https://github.com/microsoft/LightGBM/pull/4878
        expected_params["verbose_eval"] = (
            # The default value of `verbose_eval` in `lightgbm.train` has been changed to 'warn'
            # in this PR: https://github.com/microsoft/LightGBM/pull/4577
            "warn"
            if Version(lgb.__version__) > Version("3.2.1")
            else True
        )
    expected_params.update(bst_params)

    for key, val in expected_params.items():
        assert key in params
        assert params[key] == str(val)

    unlogged_params = [
        "params",
        "train_set",
        "valid_sets",
        "valid_names",
        "fobj",
        "feval",
        "init_model",
        "learning_rates",
        "callbacks",
    ]
    if Version(lgb.__version__) <= Version("3.3.1"):
        # The parameter `evals_result` in `lightgbm.train` is removed in this PR:
        # https://github.com/microsoft/LightGBM/pull/4882
        unlogged_params.append("evals_result")

    for param in unlogged_params:
        assert param not in params


def test_lgb_autolog_logs_specified_params(bst_params, train_set):
    newron.lightgbm.autolog()
    expected_params = {
        "num_boost_round": 10,
    }
    if Version(lgb.__version__) <= Version("3.3.1"):
        # The parameter `verbose_eval` in `lightgbm.train` is removed in this PR:
        # https://github.com/microsoft/LightGBM/pull/4878
        # The parameter `early_stopping_rounds` in `lightgbm.train` is removed in this PR:
        # https://github.com/microsoft/LightGBM/pull/4908
        expected_params["verbose_eval"] = False
        expected_params["early_stopping_rounds"] = 5
    lgb.train(bst_params, train_set, valid_sets=[train_set], **expected_params)
    run = get_latest_run()
    params = run.data.params

    expected_params.update(bst_params)

    for key, val in expected_params.items():
        assert key in params
        assert params[key] == str(val)

    unlogged_params = [
        "params",
        "train_set",
        "valid_sets",
        "valid_names",
        "fobj",
        "feval",
        "init_model",
        "learning_rates",
        "callbacks",
    ]
    if Version(lgb.__version__) <= Version("3.3.1"):
        # The parameter `evals_result` in `lightgbm.train` is removed in this PR:
        # https://github.com/microsoft/LightGBM/pull/4882
        unlogged_params.append("evals_result")

    for param in unlogged_params:
        assert param not in params


def test_lgb_autolog_sklearn():

    newron.lightgbm.autolog()

    X, y = datasets.load_iris(return_X_y=True)
    params = {"n_estimators": 10, "reg_lambda": 1}
    model = lgb.LGBMClassifier(**params)

    with newron.start_run() as run:
        model.fit(X, y)
        model_uri = newron.get_artifact_uri("model")

    client = NewronClient()
    run = client.get_run(run.info.run_id)
    assert run.data.metrics.items() <= params.items()
    artifacts = {x.path for x in client.list_artifacts(run.info.run_id)}
    assert artifacts >= {
        "feature_importance_gain.png",
        "feature_importance_gain.json",
        "feature_importance_split.png",
        "feature_importance_split.json",
    }
    loaded_model = newron.lightgbm.load_model(model_uri)
    np.testing.assert_allclose(loaded_model.predict(X), model.predict(X))


def test_lgb_autolog_with_sklearn_outputs_do_not_reflect_training_dataset_mutations():
    original_lgb_classifier_fit = lgb.LGBMClassifier.fit
    original_lgb_classifier_predict = lgb.LGBMClassifier.predict

    def patched_lgb_classifier_fit(self, *args, **kwargs):
        X = args[0]
        X["TESTCOL"] = 5
        return original_lgb_classifier_fit(self, *args, **kwargs)

    def patched_lgb_classifier_predict(self, *args, **kwargs):
        X = args[0]
        X["TESTCOL"] = 5
        return original_lgb_classifier_predict(self, *args, **kwargs)

    with mock.patch("lightgbm.LGBMClassifier.fit", patched_lgb_classifier_fit), mock.patch(
        "lightgbm.LGBMClassifier.predict", patched_lgb_classifier_predict
    ):
        newron.lightgbm.autolog(log_models=True, log_model_signatures=True, log_input_examples=True)

        X = pd.DataFrame(
            {
                "Total Volume": [64236.62, 54876.98, 118220.22],
                "Total Bags": [8696.87, 9505.56, 8145.35],
                "Small Bags": [8603.62, 9408.07, 8042.21],
                "Large Bags": [93.25, 97.49, 103.14],
                "XLarge Bags": [0.0, 0.0, 0.0],
            }
        )
        y = pd.Series([1, 0, 1])

        params = {"n_estimators": 10, "reg_lambda": 1}
        model = lgb.LGBMClassifier(**params)
        model.fit(X, y)

        run_artifact_uri = newron.last_active_run().info.artifact_uri
        model_conf = get_model_conf(run_artifact_uri)
        input_example = pd.read_json(
            os.path.join(run_artifact_uri, "model", "input_example.json"), orient="split"
        )
        model_signature_input_names = [inp.name for inp in model_conf.signature.inputs.inputs]
        assert "XLarge Bags" in model_signature_input_names
        assert "XLarge Bags" in input_example.columns
        assert "TESTCOL" not in model_signature_input_names
        assert "TESTCOL" not in input_example.columns


def test_lgb_autolog_logs_metrics_with_validation_data(bst_params, train_set):
    newron.lightgbm.autolog()
    evals_result = {}
    if Version(lgb.__version__) <= Version("3.3.1"):
        lgb.train(
            bst_params,
            train_set,
            num_boost_round=10,
            valid_sets=[train_set],
            valid_names=["train"],
            evals_result=evals_result,
        )
    else:
        lgb.train(
            bst_params,
            train_set,
            num_boost_round=10,
            valid_sets=[train_set],
            valid_names=["train"],
            callbacks=[lgb.record_evaluation(evals_result)],
        )
    run = get_latest_run()
    data = run.data
    client = NewronClient()
    metric_key = "train-multi_logloss"
    metric_history = [x.value for x in client.get_metric_history(run.info.run_id, metric_key)]
    assert metric_key in data.metrics
    assert len(metric_history) == 10
    assert metric_history == evals_result["train"]["multi_logloss"]


def test_lgb_autolog_logs_metrics_with_multi_validation_data(bst_params, train_set):
    newron.lightgbm.autolog()
    evals_result = {}
    # If we use [train_set, train_set] here, LightGBM ignores the first dataset.
    # To avoid that, create a new Dataset object.
    valid_sets = [train_set, lgb.Dataset(train_set.data)]
    valid_names = ["train", "valid"]
    if Version(lgb.__version__) <= Version("3.3.1"):
        lgb.train(
            bst_params,
            train_set,
            num_boost_round=10,
            valid_sets=valid_sets,
            valid_names=valid_names,
            evals_result=evals_result,
        )
    else:
        lgb.train(
            bst_params,
            train_set,
            num_boost_round=10,
            valid_sets=valid_sets,
            valid_names=valid_names,
            callbacks=[lgb.record_evaluation(evals_result)],
        )
    run = get_latest_run()
    data = run.data
    client = NewronClient()
    for valid_name in valid_names:
        metric_key = "{}-multi_logloss".format(valid_name)
        metric_history = [x.value for x in client.get_metric_history(run.info.run_id, metric_key)]
        assert metric_key in data.metrics
        assert len(metric_history) == 10
        assert metric_history == evals_result[valid_name]["multi_logloss"]


def test_lgb_autolog_logs_metrics_with_multi_metrics(bst_params, train_set):
    newron.lightgbm.autolog()
    evals_result = {}
    params = {"metric": ["multi_error", "multi_logloss"]}
    params.update(bst_params)
    valid_sets = [train_set]
    valid_names = ["train"]
    if Version(lgb.__version__) <= Version("3.3.1"):
        lgb.train(
            params,
            train_set,
            num_boost_round=10,
            valid_sets=valid_sets,
            valid_names=valid_names,
            evals_result=evals_result,
        )
    else:
        lgb.train(
            params,
            train_set,
            num_boost_round=10,
            valid_sets=valid_sets,
            valid_names=valid_names,
            callbacks=[lgb.record_evaluation(evals_result)],
        )
    run = get_latest_run()
    data = run.data
    client = NewronClient()
    for metric_name in params["metric"]:
        metric_key = "{}-{}".format(valid_names[0], metric_name)
        metric_history = [x.value for x in client.get_metric_history(run.info.run_id, metric_key)]
        assert metric_key in data.metrics
        assert len(metric_history) == 10
        assert metric_history == evals_result["train"][metric_name]


def test_lgb_autolog_logs_metrics_with_multi_validation_data_and_metrics(bst_params, train_set):
    newron.lightgbm.autolog()
    evals_result = {}
    params = {"metric": ["multi_error", "multi_logloss"]}
    params.update(bst_params)
    valid_sets = [train_set, lgb.Dataset(train_set.data)]
    valid_names = ["train", "valid"]
    if Version(lgb.__version__) <= Version("3.3.1"):
        lgb.train(
            params,
            train_set,
            num_boost_round=10,
            valid_sets=valid_sets,
            valid_names=valid_names,
            evals_result=evals_result,
        )
    else:
        lgb.train(
            params,
            train_set,
            num_boost_round=10,
            valid_sets=valid_sets,
            valid_names=valid_names,
            callbacks=[lgb.record_evaluation(evals_result)],
        )
    run = get_latest_run()
    data = run.data
    client = NewronClient()
    for valid_name in valid_names:
        for metric_name in params["metric"]:
            metric_key = "{}-{}".format(valid_name, metric_name)
            metric_history = [
                x.value for x in client.get_metric_history(run.info.run_id, metric_key)
            ]
            assert metric_key in data.metrics
            assert len(metric_history) == 10
            assert metric_history == evals_result[valid_name][metric_name]


def test_lgb_autolog_batch_metrics_logger_logs_expected_metrics(bst_params, train_set):
    patched_metrics_data = []

    # Mock patching BatchMetricsLogger.record_metrics()
    # to ensure that expected metrics are being logged.
    original = BatchMetricsLogger.record_metrics

    with patch(
        "mlflow.utils.autologging_utils.BatchMetricsLogger.record_metrics", autospec=True
    ) as record_metrics_mock:

        def record_metrics_side_effect(self, metrics, step=None):
            patched_metrics_data.extend(metrics.items())
            original(self, metrics, step)

        record_metrics_mock.side_effect = record_metrics_side_effect

        newron.lightgbm.autolog()
        evals_result = {}
        params = {"metric": ["multi_error", "multi_logloss"]}
        params.update(bst_params)
        valid_sets = [train_set, lgb.Dataset(train_set.data)]
        valid_names = ["train", "valid"]
        if Version(lgb.__version__) <= Version("3.3.1"):
            lgb.train(
                params,
                train_set,
                num_boost_round=10,
                valid_sets=valid_sets,
                valid_names=valid_names,
                evals_result=evals_result,
            )
        else:
            lgb.train(
                params,
                train_set,
                num_boost_round=10,
                valid_sets=valid_sets,
                valid_names=valid_names,
                callbacks=[lgb.record_evaluation(evals_result)],
            )

    run = get_latest_run()
    original_metrics = run.data.metrics
    patched_metrics_data = dict(patched_metrics_data)
    for metric_name in original_metrics:
        assert metric_name in patched_metrics_data
        assert original_metrics[metric_name] == patched_metrics_data[metric_name]

    assert "train-multi_logloss" in original_metrics
    assert "train-multi_logloss" in patched_metrics_data


def ranking_dataset(num_rows=100, num_queries=10):
    # https://stackoverflow.com/a/67621253
    num_rows_per_query = num_rows // num_queries
    df = pd.DataFrame(
        {
            "query_id": [i for i in range(num_queries) for _ in range(num_rows_per_query)],
            "f1": np.random.random(size=(num_rows,)),
            "f2": np.random.random(size=(num_rows,)),
            "relevance": np.random.randint(2, size=(num_rows,)),
        }
    )
    train_size = int(num_rows * 0.75)
    df_train = df[:train_size]
    df_valid = df[train_size:]
    # Train
    group_train = df_train.groupby("query_id")["query_id"].count().to_numpy()
    X_train = df_train.drop(["query_id", "relevance"], axis=1)
    y_train = df_train["relevance"]
    train_set = lgb.Dataset(X_train, y_train, group=group_train)
    # Validation
    group_val = df_valid.groupby("query_id")["query_id"].count().to_numpy()
    X_valid = df_valid.drop(["query_id", "relevance"], axis=1)
    y_valid = df_valid["relevance"]
    valid_set = lgb.Dataset(X_valid, y_valid, group=group_val)
    return train_set, valid_set


def test_lgb_autolog_atsign_metrics(train_set):
    train_set, valid_set = ranking_dataset()
    newron.lightgbm.autolog()
    params = {"objective": "regression", "metric": ["map"], "eval_at": [1]}
    lgb.train(
        params,
        train_set,
        valid_sets=[valid_set],
        valid_names=["valid"],
        num_boost_round=5,
    )
    run = get_latest_run()
    assert set(run.data.metrics) == {"valid-map_at_1"}


def test_lgb_autolog_logs_metrics_with_early_stopping(bst_params, train_set):
    newron.lightgbm.autolog()
    evals_result = {}
    params = {"metric": ["multi_error", "multi_logloss"]}
    params.update(bst_params)
    valid_sets = [train_set, lgb.Dataset(train_set.data)]
    valid_names = ["train", "valid"]
    if Version(lgb.__version__) <= Version("3.3.1"):
        model = lgb.train(
            params,
            train_set,
            num_boost_round=10,
            early_stopping_rounds=5,
            valid_sets=valid_sets,
            valid_names=valid_names,
            evals_result=evals_result,
        )
    else:
        model = lgb.train(
            params,
            train_set,
            num_boost_round=10,
            valid_sets=valid_sets,
            valid_names=valid_names,
            callbacks=[
                lgb.record_evaluation(evals_result),
                lgb.early_stopping(5),
            ],
        )
    run = get_latest_run()
    data = run.data
    client = NewronClient()
    assert "best_iteration" in data.metrics
    assert int(data.metrics["best_iteration"]) == model.best_iteration
    assert "stopped_iteration" in data.metrics
    assert int(data.metrics["stopped_iteration"]) == len(evals_result["train"]["multi_logloss"])

    for valid_name in valid_names:
        for metric_name in params["metric"]:
            metric_key = "{}-{}".format(valid_name, metric_name)
            metric_history = [
                x.value for x in client.get_metric_history(run.info.run_id, metric_key)
            ]
            assert metric_key in data.metrics

            best_metrics = evals_result[valid_name][metric_name][model.best_iteration - 1]
            assert metric_history == evals_result[valid_name][metric_name] + [best_metrics]


def test_lgb_autolog_logs_feature_importance(bst_params, train_set):
    newron.lightgbm.autolog()
    model = lgb.train(bst_params, train_set, num_boost_round=10)
    run = get_latest_run()
    run_id = run.info.run_id
    artifacts_dir = run.info.artifact_uri.replace("file://", "")
    client = NewronClient()
    artifacts = [x.path for x in client.list_artifacts(run_id)]

    for imp_type in ["split", "gain"]:
        plot_name = "feature_importance_{}.png".format(imp_type)
        assert plot_name in artifacts

        json_name = "feature_importance_{}.json".format(imp_type)
        assert json_name in artifacts

        json_path = os.path.join(artifacts_dir, json_name)
        with open(json_path, "r") as f:
            loaded_imp = json.load(f)

        features = model.feature_name()
        importance = model.feature_importance(importance_type=imp_type)
        imp = dict(zip(features, importance.tolist()))

        assert loaded_imp == imp


def test_no_figure_is_opened_after_logging(bst_params, train_set):
    newron.lightgbm.autolog()
    lgb.train(bst_params, train_set, num_boost_round=10)
    assert mpl.pyplot.get_fignums() == []


def test_lgb_autolog_loads_model_from_artifact(bst_params, train_set):
    newron.lightgbm.autolog()
    model = lgb.train(bst_params, train_set, num_boost_round=10)
    run = get_latest_run()
    run_id = run.info.run_id

    loaded_model = newron.lightgbm.load_model("runs:/{}/model".format(run_id))
    np.testing.assert_array_almost_equal(
        model.predict(train_set.data), loaded_model.predict(train_set.data)
    )


def test_lgb_autolog_gets_input_example(bst_params):
    # we need to check the example input against the initial input given to train function.
    # we can't use the train_set fixture for this as it defines free_raw_data=False but this
    # feature should work even if it is True
    iris = datasets.load_iris()
    X = pd.DataFrame(iris.data[:, :2], columns=iris.feature_names[:2])
    y = iris.target
    dataset = lgb.Dataset(X, y, free_raw_data=True)

    newron.lightgbm.autolog(log_input_examples=True)
    lgb.train(bst_params, dataset)
    run = get_latest_run()

    model_path = os.path.join(run.info.artifact_uri, "model")
    model_conf = Model.load(os.path.join(model_path, "MLmodel"))

    input_example = _read_example(model_conf, model_path)

    pd.testing.assert_frame_equal(input_example, (X[:5]))

    pyfunc_model = newron.pyfunc.load_model(os.path.join(run.info.artifact_uri, "model"))

    # make sure reloading the input_example and predicting on it does not error
    pyfunc_model.predict(input_example)


def test_lgb_autolog_infers_model_signature_correctly(bst_params):
    iris = datasets.load_iris()
    X = pd.DataFrame(iris.data[:, :2], columns=iris.feature_names[:2])
    y = iris.target
    dataset = lgb.Dataset(X, y, free_raw_data=True)

    newron.lightgbm.autolog(log_model_signatures=True)
    lgb.train(bst_params, dataset)
    run = get_latest_run()
    run_id = run.info.run_id
    artifacts_dir = run.info.artifact_uri.replace("file://", "")
    client = NewronClient()
    artifacts = [x.path for x in client.list_artifacts(run_id, "model")]

    ml_model_filename = "MLmodel"
    assert str(os.path.join("model", ml_model_filename)) in artifacts
    ml_model_path = os.path.join(artifacts_dir, "model", ml_model_filename)

    data = None
    with open(ml_model_path, "r") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)

    assert data is not None
    assert "signature" in data
    signature = data["signature"]
    assert signature is not None

    assert "inputs" in signature
    assert json.loads(signature["inputs"]) == [
        {"name": "sepal length (cm)", "type": "double"},
        {"name": "sepal width (cm)", "type": "double"},
    ]

    assert "outputs" in signature
    assert json.loads(signature["outputs"]) == [
        {"type": "tensor", "tensor-spec": {"dtype": "float64", "shape": [-1, 3]}},
    ]


def test_lgb_autolog_continues_logging_even_if_signature_inference_fails(tmpdir):
    tmp_csv = tmpdir.join("data.csv")
    tmp_csv.write("2,6.4,2.8,5.6,2.2\n")
    tmp_csv.write("1,5.0,2.3,3.3,1.0\n")
    tmp_csv.write("2,4.9,2.5,4.5,1.7\n")
    tmp_csv.write("0,4.9,3.1,1.5,0.1\n")
    tmp_csv.write("0,5.7,3.8,1.7,0.3\n")

    # signature and input example inference should fail here since the dataset is given
    #   as a file path
    dataset = lgb.Dataset(tmp_csv.strpath)

    bst_params = {
        "objective": "multiclass",
        "num_class": 3,
    }

    newron.lightgbm.autolog(log_model_signatures=True)
    lgb.train(bst_params, dataset)
    run = get_latest_run()
    run_id = run.info.run_id
    artifacts_dir = run.info.artifact_uri.replace("file://", "")
    client = NewronClient()
    artifacts = [x.path for x in client.list_artifacts(run_id, "model")]

    ml_model_filename = "MLmodel"
    assert os.path.join("model", ml_model_filename) in artifacts
    ml_model_path = os.path.join(artifacts_dir, "model", ml_model_filename)

    data = None
    with open(ml_model_path, "r") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)

    assert data is not None
    assert "run_id" in data
    assert "signature" not in data


@pytest.mark.parametrize("log_input_examples", [True, False])
@pytest.mark.parametrize("log_model_signatures", [True, False])
def test_lgb_autolog_configuration_options(bst_params, log_input_examples, log_model_signatures):
    iris = datasets.load_iris()
    X = pd.DataFrame(iris.data[:, :2], columns=iris.feature_names[:2])
    y = iris.target

    with newron.start_run() as run:
        newron.lightgbm.autolog(
            log_input_examples=log_input_examples, log_model_signatures=log_model_signatures
        )
        dataset = lgb.Dataset(X, y)
        lgb.train(bst_params, dataset)
    model_conf = get_model_conf(run.info.artifact_uri)
    assert ("saved_input_example_info" in model_conf.to_dict()) == log_input_examples
    assert ("signature" in model_conf.to_dict()) == log_model_signatures


@pytest.mark.parametrize("log_models", [True, False])
def test_lgb_autolog_log_models_configuration(bst_params, log_models):
    iris = datasets.load_iris()
    X = pd.DataFrame(iris.data[:, :2], columns=iris.feature_names[:2])
    y = iris.target

    with newron.start_run() as run:
        newron.lightgbm.autolog(log_models=log_models)
        dataset = lgb.Dataset(X, y)
        lgb.train(bst_params, dataset)

    run_id = run.info.run_id
    client = NewronClient()
    artifacts = [f.path for f in client.list_artifacts(run_id)]
    assert ("model" in artifacts) == log_models


def test_lgb_autolog_does_not_break_dataset_instantiation_with_data_none():
    """
    This test verifies that `lightgbm.Dataset(None)` doesn't fail after patching.
    LightGBM internally calls `lightgbm.Dataset(None)` to create a subset of `Dataset`:
    https://github.com/microsoft/LightGBM/blob/v3.0.0/python-package/lightgbm/basic.py#L1381
    """
    newron.lightgbm.autolog()
    lgb.Dataset(None)


def test_callback_func_is_pickable():
    cb = picklable_exception_safe_function(
        functools.partial(_autolog_callback, BatchMetricsLogger(run_id="1234"), eval_results={})
    )
    pickle.dumps(cb)


def test_sklearn_api_autolog_registering_model():
    registered_model_name = "test_autolog_registered_model"
    newron.lightgbm.autolog(registered_model_name=registered_model_name)

    X, y = datasets.load_iris(return_X_y=True)
    params = {"n_estimators": 10, "reg_lambda": 1}
    model = lgb.LGBMClassifier(**params)

    with newron.start_run():
        model.fit(X, y)

        registered_model = NewronClient().get_registered_model(registered_model_name)
        assert registered_model.name == registered_model_name


def test_lgb_api_autolog_registering_model(bst_params, train_set):
    registered_model_name = "test_autolog_registered_model"
    newron.lightgbm.autolog(registered_model_name=registered_model_name)

    with newron.start_run():
        lgb.train(bst_params, train_set, num_boost_round=1)

        registered_model = NewronClient().get_registered_model(registered_model_name)
        assert registered_model.name == registered_model_name


import os
import tempfile
import pandas as pd
from pathlib import Path
import pytest
from pyspark.sql import SparkSession
from pyspark.sql.utils import AnalysisException
from sklearn.datasets import load_diabetes
from unittest import mock

from mlflow.exceptions import MlflowException
from mlflow.pipelines.utils import _PIPELINE_CONFIG_FILE_NAME
from mlflow.pipelines.steps.predict import PredictStep, _SCORED_OUTPUT_FILE_NAME

# pylint: disable=unused-import
from tests.pipelines.helper_functions import (
    enter_test_pipeline_directory,
    enter_pipeline_example_directory,
    get_random_id,
    tmp_pipeline_exec_path,
    tmp_pipeline_root_path,
    train_and_log_model,
    train_log_and_register_model,
)  # pylint: enable=unused-import


@pytest.fixture(scope="module", autouse=True)
def spark_session():
    spark_warehouse_path = os.path.abspath(tempfile.mkdtemp())
    session = (
        SparkSession.builder.master("local[*]")
        .config("spark.jars.packages", "io.delta:delta-core_2.12:1.2.1")
        .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
        .config(
            "spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog"
        )
        .config("spark.sql.warehouse.dir", str(spark_warehouse_path))
        .getOrCreate()
    )
    yield session
    session.stop()


@pytest.fixture(autouse=True)
def patch_env_manager():
    with mock.patch("mlflow.pipelines.steps.predict._ENV_MANAGER", "local"):
        yield


def prediction_assertions(output_dir: Path, output_format: str, output_name: str, spark):
    if output_format == "table":
        sdf = spark.table(output_name)
        df = sdf.toPandas()
    else:
        file_name = "{}.{}".format(output_name, output_format)
        (output_dir / file_name).exists()
        df = pd.read_parquet(output_dir / file_name)
    assert "prediction" in df
    assert (df.prediction == 42).all()


# Sets up predict step run and returns output directory
@pytest.fixture(autouse=True)
def predict_step_output_dir(tmp_pipeline_root_path: Path, tmp_pipeline_exec_path: Path):
    ingest_scoring_step_output_dir = tmp_pipeline_exec_path.joinpath(
        "steps", "ingest_scoring", "outputs"
    )
    ingest_scoring_step_output_dir.mkdir(parents=True)
    X, _ = load_diabetes(as_frame=True, return_X_y=True)
    X.to_parquet(ingest_scoring_step_output_dir.joinpath("scoring-dataset.parquet"))
    predict_step_output_dir = tmp_pipeline_exec_path.joinpath("steps", "predict", "outputs")
    predict_step_output_dir.mkdir(parents=True)
    pipeline_yaml = tmp_pipeline_root_path.joinpath(_PIPELINE_CONFIG_FILE_NAME)
    pipeline_yaml.write_text(
        """
template: "{template_to_use}"
""".format(
            template_to_use="regression/v1",
        )
    )
    return predict_step_output_dir


@pytest.mark.parametrize("register_model", [True, False])
def test_predict_step_runs(
    tmp_pipeline_root_path: Path,
    predict_step_output_dir: Path,
    spark_session,
    register_model: bool,
):
    if register_model:
        model_name = "model_" + get_random_id()
        model_uri = train_log_and_register_model(model_name, is_dummy=True)
    else:
        run_id, _ = train_and_log_model(is_dummy=True)
        model_uri = "runs:/{run_id}/{artifact_path}".format(
            run_id=run_id, artifact_path="train/model"
        )

    predict_step = PredictStep.from_pipeline_config(
        {
            "steps": {
                "predict": {
                    "model_uri": model_uri,
                    "output_format": "parquet",
                    "output_location": str(predict_step_output_dir.joinpath("output.parquet")),
                }
            }
        },
        str(tmp_pipeline_root_path),
    )
    predict_step._run(str(predict_step_output_dir))

    # Test internal predict step output artifact
    artifact_file_name, artifact_file_extension = _SCORED_OUTPUT_FILE_NAME.split(".")
    prediction_assertions(
        predict_step_output_dir, artifact_file_extension, artifact_file_name, spark_session
    )
    # Test user specified output
    prediction_assertions(predict_step_output_dir, "parquet", "output", spark_session)


def test_predict_step_uses_register_step_model_name(
    tmp_pipeline_root_path: Path,
    predict_step_output_dir: Path,
    spark_session,
):
    rm_name = "register_step_model"
    train_log_and_register_model(rm_name, is_dummy=True)

    predict_step = PredictStep.from_pipeline_config(
        {
            "steps": {
                "register": {"model_name": rm_name},
                "predict": {
                    "output_format": "parquet",
                    "output_location": str(predict_step_output_dir.joinpath("output.parquet")),
                },
            }
        },
        str(tmp_pipeline_root_path),
    )
    predict_step._run(str(predict_step_output_dir))

    prediction_assertions(predict_step_output_dir, "parquet", "output", spark_session)


def test_predict_model_uri_takes_precendence_over_model_name(
    tmp_pipeline_root_path: Path,
    predict_step_output_dir: Path,
    spark_session,
):
    # Train a normal model and a dummy model
    register_step_rm_name = "register_step_model"
    train_log_and_register_model(register_step_rm_name)
    model_uri = train_log_and_register_model("predict_step_model", is_dummy=True)

    # Specify the normal model in the register step `model_name` config key and
    # the dummy model in the predict step `model_uri` config key
    predict_step = PredictStep.from_pipeline_config(
        {
            "steps": {
                "register": {"model_name": register_step_rm_name},
                "predict": {
                    "model_uri": model_uri,
                    "output_format": "parquet",
                    "output_location": str(predict_step_output_dir.joinpath("output.parquet")),
                },
            }
        },
        str(tmp_pipeline_root_path),
    )
    predict_step._run(str(predict_step_output_dir))

    # These assertions will only pass if the dummy model was used for scoring
    prediction_assertions(predict_step_output_dir, "parquet", "output", spark_session)


@pytest.mark.parametrize("output_format", ["parquet", "delta", "table"])
def test_predict_step_output_formats(
    tmp_pipeline_root_path: Path, predict_step_output_dir: Path, spark_session, output_format: str
):
    rm_name = "model_" + get_random_id()
    output_name = "output_" + get_random_id()
    model_uri = train_log_and_register_model(rm_name, is_dummy=True)

    pipeline_config = {
        "steps": {
            "predict": {
                "model_uri": model_uri,
                "output_format": output_format,
            }
        }
    }
    if output_format == "table":
        pipeline_config["steps"]["predict"]["output_location"] = output_name
    else:
        file_name = "{}.{}".format(output_name, output_format)
        pipeline_config["steps"]["predict"]["output_location"] = str(
            predict_step_output_dir / file_name
        )
    predict_step = PredictStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    predict_step._run(str(predict_step_output_dir))
    prediction_assertions(predict_step_output_dir, output_format, output_name, spark_session)


@pytest.mark.parametrize("output_format", ["parquet", "delta", "table"])
def test_predict_throws_when_overwriting_data(
    tmp_pipeline_root_path: Path, predict_step_output_dir: Path, spark_session, output_format: str
):
    rm_name = "model_" + get_random_id()
    model_uri = train_log_and_register_model(rm_name, is_dummy=True)
    sdf = spark_session.createDataFrame(
        [
            (0, "a b c d e spark", 1.0),
            (1, "b d", 0.0),
            (2, "spark f g h", 1.0),
            (3, "hadoop mapreduce", 0.0),
        ],
        ["id", "text", "label"],
    )
    if output_format == "table":
        output_path = get_random_id()
        sdf.write.format("delta").saveAsTable(output_path)
    else:
        output_file = "output_{}.{}".format(get_random_id(), output_format)
        output_path = str(predict_step_output_dir / output_file)
        sdf.coalesce(1).write.format(output_format).save(output_path)

    pipeline_config = {
        "steps": {
            "predict": {
                "model_uri": model_uri,
                "output_format": output_format,
                "output_location": output_path,
            }
        }
    }

    predict_step = PredictStep.from_pipeline_config(pipeline_config, str(tmp_pipeline_root_path))
    with pytest.raises(AnalysisException, match="already exists"):
        predict_step._run(str(predict_step_output_dir))


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_predict_throws_when_improperly_configured():
    with pytest.raises(MlflowException, match="Config for predict step is not found"):
        PredictStep.from_pipeline_config(
            pipeline_config={},
            pipeline_root=os.getcwd(),
        )

    for required_key in ["output_format", "output_location"]:
        pipeline_config = {
            "steps": {
                "predict": {
                    "model_uri": "models:/taxi_fare_regressor/Production",
                    "output_format": "parquet",
                    "output_location": "random/path",
                }
            }
        }
        pipeline_config["steps"]["predict"].pop(required_key)
        with pytest.raises(
            MlflowException, match=f"The `{required_key}` configuration key must be specified"
        ):
            PredictStep.from_pipeline_config(
                pipeline_config=pipeline_config,
                pipeline_root=os.getcwd(),
            )

    with pytest.raises(
        MlflowException, match="Invalid `output_format` in predict step configuration"
    ):
        PredictStep.from_pipeline_config(
            pipeline_config={
                "steps": {
                    "predict": {
                        "model_uri": "my model",
                        "output_format": "fancy_format",
                        "output_location": "random/path",
                    }
                }
            },
            pipeline_root=os.getcwd(),
        )


@pytest.mark.usefixtures("enter_test_pipeline_directory")
def test_predict_throws_when_no_model_is_specified():
    pipeline_config = {
        "steps": {
            "predict": {
                "output_format": "parquet",
                "output_location": "random/path",
            }
        }
    }
    with pytest.raises(MlflowException, match="No model specified for batch scoring"):
        PredictStep.from_pipeline_config(
            pipeline_config=pipeline_config,
            pipeline_root=os.getcwd(),
        )


def test_predict_skips_profiling_when_specified(
    tmp_pipeline_root_path: Path,
    predict_step_output_dir: Path,
):
    model_name = "model_" + get_random_id()
    model_uri = train_log_and_register_model(model_name, is_dummy=True)
    with mock.patch("mlflow.pipelines.utils.step.get_pandas_data_profile") as mock_profiling:
        PredictStep.from_pipeline_config(
            {
                "steps": {
                    "predict": {
                        "model_uri": model_uri,
                        "output_format": "parquet",
                        "output_location": str(predict_step_output_dir.joinpath("output.parquet")),
                        "skip_data_profiling": True,
                    }
                }
            },
            str(tmp_pipeline_root_path),
        ).run(str(predict_step_output_dir))

    expected_step_card_path = os.path.join(str(predict_step_output_dir), "card.html")
    with open(expected_step_card_path, "r") as f:
        step_card_html_content = f.read()
    assert "Profile of Scored Dataset" not in step_card_html_content
    mock_profiling.assert_not_called()



# pylint: disable=wrong-import-position
"""
The ``newron`` module provides a high-level API for starting and managing Newron runs.
For example:

.. code:: python

    import newron

    newron.init()

    newron.start_run()
    newron.log_param("my", "param")
    newron.log_metric("score", 100)
    newron.end_run()

You can also use the context manager syntax like this:

.. code:: python

    import newron

    newron.init()
    
    with newron.start_run() as run:
        newron.log_param("my", "param")
        newron.log_metric("score", 100)

which automatically terminates the run at the end of the ``with`` block.
"""
from newron.version import VERSION as __version__  # pylint: disable=unused-import
from mlflow.utils.logging_utils import _configure_mlflow_loggers


# Filter annoying Cython warnings that serve no good purpose, and so before
# importing other modules.
# See: https://github.com/numpy/numpy/pull/432/commits/170ed4e33d6196d7
import warnings

warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

from newron import tracking
from newron import models
from newron import auth
from mlflow import projects
import newron.artifacts

# model flavors
_model_flavors_supported = []
    # pylint: disable=unused-import
from newron import catboost
from newron import fastai
from newron import gluon
from newron import keras
from newron import lightgbm
from mlflow import pyfunc
from newron import pytorch
from newron import sklearn
from newron import spacy
from newron import tensorflow
from mlflow import xgboost
from mlflow import shap
from mlflow import pyspark
from mlflow import paddle
from newron import prophet
from newron import diviner

_model_flavors_supported = [
    "catboost",
    "fastai",
    "gluon",
    "keras",
    "lightgbm",
    "pyfunc",
    "pytorch",
    "sklearn",
    "spacy",
    "spark",
    "tensorflow",
    "xgboost",
    "shap",
    "paddle",
    "prophet",
    "diviner",
]

_configure_mlflow_loggers(root_module_name=__name__)

# TODO: Comment out this block when we deprecate support for python 3.7.
# _major = 3
# _minor = 7
# _deprecated_version = (_major, _minor)
# _min_supported_version = (_major, _minor + 1)

# if sys.version_info[:2] == _deprecated_version:
#     warnings.warn(
#         "MLflow support for Python {dep_ver} is deprecated and will be dropped in "
#         "an upcoming release. At that point, existing Python {dep_ver} workflows "
#         "that use MLflow will continue to work without modification, but Python {dep_ver} "
#         "users will no longer get access to the latest MLflow features and bugfixes. "
#         "We recommend that you upgrade to Python {min_ver} or newer.".format(
#             dep_ver=".".join(map(str, _deprecated_version)),
#             min_ver=".".join(map(str, _min_supported_version)),
#         ),
#         FutureWarning,
#         stacklevel=2,
#     )


#search_experiments = mlflow.tracking.search_experiments 
#Marked as experimental

get_tracking_uri = tracking.get_tracking_uri
get_registry_uri = tracking.get_registry_uri
create_experiment = tracking.create_experiment
set_experiment = tracking.set_experiment
log_params = tracking.log_params
log_metrics = tracking.log_metrics
set_experiment_tags = tracking.set_experiment_tags
set_experiment_tag = tracking.set_experiment_tag
set_tags = tracking.set_tags
delete_experiment = tracking.delete_experiment  
delete_run = tracking.delete_run
register_model = tracking.register_model
autolog = tracking.autolog
evaluate = models.evaluate
last_active_run = tracking.last_active_run
NewronClient = tracking.NewronClient
ActiveRun = tracking.ActiveRun
log_param = tracking.log_param
log_metric = tracking.log_metric
set_tag = tracking.set_tag
delete_tag = tracking.delete_tag
log_artifacts = tracking.log_artifacts
log_artifact = tracking.log_artifact
log_text = tracking.log_text
log_dict = tracking.log_dict
log_image = tracking.log_image
log_figure = tracking.log_figure
active_run = tracking.active_run
get_run = tracking.get_run
start_run = tracking.start_run
end_run = tracking.end_run
search_runs = tracking.search_runs
list_run_infos = tracking.list_run_infos
get_artifact_uri = tracking.get_artifact_uri
set_tracking_uri = tracking.set_tracking_uri
set_registry_uri = tracking.set_registry_uri
get_experiment = tracking.get_experiment
get_experiment_by_name = tracking.get_experiment_by_name
list_experiments = tracking.list_experiments
run = projects.run
init = tracking.init


__all__ = [
    "ActiveRun",
    "log_param",
    "log_params",
    "log_metric",
    "log_metrics",
    "set_experiment_tags",
    "set_experiment_tag",
    "set_tag",
    "set_tags",
    "delete_tag",
    "log_artifacts",
    "log_artifact",
    "log_text",
    "log_dict",
    "log_figure",
    "log_image",
    "active_run",
    "start_run",
    "end_run",
    "search_runs",
    "get_artifact_uri",
    "get_tracking_uri",
    "set_tracking_uri",
    "get_experiment",
    "get_experiment_by_name",
    "list_experiments",
    "search_experiments",
    "create_experiment",
    "set_experiment",
    "delete_experiment",
    "get_run",
    "delete_run",
    "run",
    "register_model",
    "get_registry_uri",
    "set_registry_uri",
    "list_run_infos",
    "autolog",
    "evaluate",
    "last_active_run",
    "NewronClient",
] + _model_flavors_supported
